#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <termios.h>
#include <time.h>

#include <dirent.h>

#include "ambiente.h"

#define FILE_NAME_MAX_SIZE 20

int alphasort(const void *, const void *);
int scandir(const char *, struct dirent ***, int(*)(const struct dirent *), int (*)(const void *, const void *));
void mostraInformacao(char *arquivo);

int pegaTecla( ) {
   struct termios oldt,  newt;
   int            ch;
   tcgetattr( STDIN_FILENO, &oldt );
   newt = oldt;
   newt.c_lflag &= ~( ICANON | ECHO );
   tcsetattr( STDIN_FILENO, TCSANOW, &newt );
   ch = getchar();
   tcsetattr( STDIN_FILENO, TCSANOW, &oldt );
   return ch;
}

/*****************************************
  Funcoes que carrega e limpa o ambiente
  *****************************************/
ambiente novoAmbiente() {
   ambiente ambiente;
   ambiente = malloc(sizeof*ambiente);
   if(ambiente == NULL)
      return NULL;

   ambiente->quantidade_arquivos = scandir(".", &ambiente->arquivos, 0, alphasort);
   if( ambiente->arquivos == NULL ) {
      free(ambiente);
      return NULL;
   }

   ambiente->quantidade_pagina = 10;
   ambiente->posicao_pagina = 0;
   ambiente->posicao_indicador = 0;

   getcwd(&ambiente->path[0], (size_t) 999);


   return ambiente;
}

void freeAmbiente(ambiente ambiente) {
   int i;
   for(i=0;i<ambiente->quantidade_arquivos;i++) {
      free(ambiente->arquivos[i]);
   }
   free(ambiente->arquivos);
   free(ambiente);

}


/*****************************************
  Funcoes para mostrar informacoes na tela
  *****************************************/

void limpaTela() {

   char esc = 27;
   fprintf(stdout, "%c%s", esc, "[2J");
   fprintf(stdout, "%c%s", esc, "[1;1H");
   fflush(stdout);

}

   void imprimeEspacos(int qtd) {
      while(qtd-- > 0)
	 putchar(' ');  
   }


/* mostra uma linha de arquivo */
void mostraLinha(char *nome, int posicao, int ativo) {
   printf("%s %3d. %s ", 
	 ativo ? "->" : "  ",
	 posicao,
	 nome);
   imprimeEspacos( FILE_NAME_MAX_SIZE - strlen(nome));
   mostraInformacao(nome);
}


void listaArquivos(ambiente ambiente) {
   int i;

   limpaTela();
   printf("Diretorio Corrente: %s\n", ambiente->path);
   for(i=0;i<ambiente->quantidade_pagina && (i + ambiente->posicao_pagina) < ambiente->quantidade_arquivos;i++) {

      mostraLinha(ambiente->arquivos[ambiente->posicao_pagina+i]->d_name, ambiente->posicao_pagina + i, ambiente->posicao_indicador == (ambiente->posicao_pagina + i));
   }
   printf("V=Visualizar, E=Editar, R=Executar, C=Mudar Dir, U=Subir, D=Descer, X=Sair\n");
   fflush(NULL);
}


/* mostra informacoes detalhadas de cada arquivo, usando a stat */
void mostraInformacao(char *arquivo) {
   struct stat statArq;
   char data[50], letra = ' ' ;
   int mode;
   off_t size;

   struct tm * time;

   stat(arquivo, &statArq);

   time = localtime(&statArq.st_mtime);

   /* DATA */
   strftime( data, 49, "%d %b %y", time);
   printf("%s ", data);

   /* tamanho */
   size = statArq.st_size;
   if( size >= 1000 ) {
      size = (size + 50)/100;
      letra = 'K';
      if( size >= 1000 ) {
	 size = (size + 500)/1000;
	 letra = 'M';
      }
   }

   if( letra == ' ') {
      printf("%7d ", (int) size );
   } else {
      printf("%4d.%d%c ", (int) (size/10), (int) (size % 10), letra);
   }

   /* permissoes */

   mode = statArq.st_mode;
   printf("%c%c%c%c%c%c%c%c%c%c",
	 /* diretorio */
	 S_ISDIR(mode) ? 'd' : '-', 

	 /* u */
	 (1<<8 & mode) ? 'r' : '-',
	 (1<<7 & mode) ? 'w' : '-',
	 (1<<6 & mode) ? 'x' : '-',

	 /* g */
	 (1<<5 & mode) ? 'r' : '-',
	 (1<<4 & mode) ? 'w' : '-',
	 (1<<3 & mode) ? 'x' : '-',

	 /* o */
	 (1<<2 & mode) ? 'r' : '-',
	 (1<<1 & mode) ? 'w' : '-',
	 (1 & mode) ? 'x' : '-');


   printf("\n");

}

/***********************************************
  Funcoes de verificacao
 ***********************************************/
int ehDiretorio(char *arquivo) {
   struct stat statArq;
   int mode;
   stat(arquivo, &statArq);

   mode = (int) statArq.st_mode;
   return S_ISDIR(mode);
}

int ehExecutavel(char *arquivo) {
   struct stat statArq;
   int mode;

   stat(arquivo, &statArq);

   mode = (int)statArq.st_mode;
   return (mode%2 == 1) && !(S_ISDIR(mode));
}

/*******************************************
  Funcoes para os comandos possiveis do programa
 *******************************************/

ambiente comandoVisualizar(ambiente ambiente) {
   int pid, status;
   char *parametros[3];



   if( (pid = fork()) < 0) {
      printf("Erro no fork\n");
      exit(1);
   }

   if(pid == 0) {
      parametros[0] = "less";
      parametros[1] = ambiente->arquivos[ambiente->posicao_indicador]->d_name;
      parametros[2] = NULL;

      execvp("less", parametros);

   } else {
      waitpid(pid, &status, 0);
   }
   return ambiente;
}
ambiente comandoEditar(ambiente ambiente) {
   int pid, status;
   char *parametros[3];



   if( (pid = fork()) < 0) {
      printf("Erro no fork\n");
      exit(1);
   }

   if(pid == 0) {
      parametros[0] = "vim";
      parametros[1] = ambiente->arquivos[ambiente->posicao_indicador]->d_name;
      parametros[2] = NULL;

      execvp("vim", parametros);

   } else {
      waitpid(pid, &status, 0);
   }

   return ambiente;
}

ambiente comandoExecutar(ambiente ambiente) {
   int pid, status, i, ct;
   char linhaParametros[5000];
   char *parametros[100];

   if( !ehExecutavel(ambiente->arquivos[ambiente->posicao_indicador]->d_name) ) {
      printf("Arquivo sem permissao para executar. Aperte qualquer tecla para continuar.\n");
      pegaTecla();
      return ambiente;
   }
   printf("Digite os parametros: ");
   fgets(linhaParametros, 4999, stdin);
   ct = 2;
   parametros[0] = ambiente->arquivos[ambiente->posicao_indicador]->d_name;
   parametros[1] = &linhaParametros[0];

   for(i =0; linhaParametros[i] != '\0'; i++) {
      if(linhaParametros[i] == ' ') {
	 parametros[ct] = &linhaParametros[i+1];
	 ct++;
	 linhaParametros[i] = '\0';	 
      }
   }
   parametros[ct] = NULL;

   if( (pid = fork()) < 0) {
      printf("Erro no fork\n");
      exit(1);
   }

   if(pid == 0) {

      execvp(parametros[0], parametros);

   } else {
      waitpid(pid, &status, 0);
   }

   return ambiente;


}

ambiente comandoMudarDir(ambiente ambiente) {
   if( !ehDiretorio(ambiente->arquivos[ambiente->posicao_indicador]->d_name) ) {
      printf("Nao eh um diretorio. Aperte qualquer tecla para continuar.\n");
      pegaTecla();
      return ambiente;
   }
   chdir(ambiente->arquivos[ambiente->posicao_indicador]->d_name);
   freeAmbiente(ambiente);
   return novoAmbiente();
}

/* main, com o loop que faz o programa rodar */
int main() {
   ambiente ambiente = novoAmbiente(); 
   int comando = 'A';
   while(1) {
      listaArquivos(ambiente);      
      comando = pegaTecla();
      switch(comando) {
	 case 'u':
	 case 'U':
	    ambiente->posicao_indicador -=1;
	    if(ambiente->posicao_indicador < 0)
	       ambiente->posicao_indicador = 0;
	    if(ambiente->posicao_indicador < ambiente->posicao_pagina)
	       ambiente->posicao_pagina -=1;
	    break;
	 case 'd':
	 case 'D':
	    ambiente->posicao_indicador +=1;
	    if(ambiente->posicao_indicador >= ambiente->quantidade_arquivos)
	       ambiente->posicao_indicador = ambiente->quantidade_arquivos -1;
	    if(ambiente->posicao_indicador >= (ambiente->posicao_pagina + ambiente->quantidade_pagina))
	       ambiente->posicao_pagina +=1;
	    break;
	 case 'x':
	 case 'X':
	    exit(0);
	    break;
	 case 'V':
	 case 'v':
	    comandoVisualizar(ambiente);
	    break;
	 case 'E':
	 case 'e':
	    comandoEditar(ambiente);
	    break;
	 case 'R':
	 case 'r':
	    comandoExecutar(ambiente);
	    break;
	 case 'C':
	 case 'c':
	    comandoMudarDir(ambiente);
	    break;
	 default:
	    printf("Comando %c nao existe, aperte qualquer tecla apra continuar\n", comando);
	    pegaTecla();
      }
   }
   return 0;
}
